struct SSeries
{
    string mName;
    double mMean;
    double mPoints[];
};

SSeries iMultipliers[], iRotations[], iResults[], iLowerMatrix[], iUpperMatrix[], iMinors[];

void SetupPCA(const int rows, const int cols)
{
    SetupMatrix(iMinors, rows, cols);
    SetupMatrix(iResults, rows, cols);
    SetupMatrix(iRotations, rows, cols);
    SetupMatrix(iMultipliers, rows, cols);
    SetupMatrix(iLowerMatrix, rows, cols);
    SetupMatrix(iUpperMatrix, rows, cols);
}

void SetupMatrix(SSeries& series[], const int rows, const int cols)
{
    ArrayResize(series, rows);
    
    for (int k = 0; k < rows; k++)
    {
        ArrayResize(series[k].mPoints, cols);
    }
}

void ClearMatrix(SSeries& series[], const int rows, const int cols)
{
    for (int k = 0; k < rows; k++)
    {
        ArrayFill(series[k].mPoints, 0, cols, 0);
    }
}

void CopyMatrix(SSeries& source[], SSeries& destination[], const int rows, const int cols)
{
    for (int k = 0; k < rows; k++)
    {
        for (int n = 0; n < cols; n++)
        {
            destination[k].mPoints[n] = source[k].mPoints[n];
        }
    }
}

double GetBarByTime(const string symbol, const ENUM_TIMEFRAMES period, const datetime time)
{
    return Bars(symbol, period, time, SeriesInfoInteger(symbol, period, SERIES_LASTBAR_DATE)) - 1;
}

int GetMinBars(SSeries& series[], const ENUM_TIMEFRAMES period, const int count)
{
    int min = 1000000;

    for (int k = 0; k < count; k++)
    {
        int bars = Bars(series[k].mName, period);
        min = min < bars ? min : bars;
    }

    return(min);
}

datetime GetLastTime(SSeries& series[], const ENUM_TIMEFRAMES period, const int count)
{
    datetime startTime = 0;

    for (int k = 0; k < count; k++)
    {
        datetime times[];
        CopyTime(series[k].mName, period, 0, 1, times);
        startTime = startTime > times[0] ? startTime : times[0];
    }

    return(startTime);
}

int Synchronize(SSeries& series[], const ENUM_TIMEFRAMES period, const int symbols, const int depth)
{
    int seconds = PeriodSeconds(period);
    datetime lastTime = GetLastTime(series, period, symbols);

    for (int k = 0; k < symbols; k++)
    {
        double closes[];
        datetime currentTime = lastTime;
        int copied = CopyClose(series[k].mName, period, 0, depth, closes);

        if (copied > 0)
        {
            int closeIndex = ArraySize(closes) - 1;
            series[k].mPoints[depth - 1] = closes[closeIndex];

            for (int n = depth - 1; n >= 0; n--)
            {
                if (GetBarByTime(series[k].mName, period, currentTime) > 0 && closeIndex > 0)
                {
                    closeIndex--;
                }

                series[k].mPoints[n] = closes[closeIndex];
                currentTime -= seconds;
            }
        }
        else
        {
            CustomLog("Synchronization : " + series[k].mName);
            return(0);
        }
    }

    CustomLog(" ");
    return(1);
}

string GetSymbolNames(const string names, const string base)
{
    string symbols = names;

    if (StringLen(names) < 1)
    {
        int count = SymbolsTotal(false);
        
        for (int k = 0; k < count; k++)
        {
            string symbol = SymbolName(k, false);
            
            if (StringFind(symbol, base) > -1)
            {
                symbols += symbol + (k == count ? "" : ",");
            }
        }
    }

    return(symbols);
}

int ListSymbols(SSeries& series[], const string names)
{
    string symbols[];
    int index = 0, count = StringSplit(names, ',', symbols);

    for (int k = 0; k < count; k++)
    {
        if (SymbolSelect(symbols[k], true) && SymbolInfoInteger(symbols[k], SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL)
        {
            ArrayResize(series, index + 1); 
            series[index++].mName = symbols[k];
        }
    }
    
    return(index);
}

void GetBaseMatrix(SSeries& series[], SSeries& base[], const int order, const int depth)
{
    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < depth; n++)
        {
            base[k].mPoints[n] = 1000 * MathLog(series[k].mPoints[n]);
        }
    }
}

void GetCovMatrix(SSeries& series[], SSeries& covs[], const int order, const int depth)
{
    for (int k = 0; k < order; k++)
    {
        series[k].mMean = 0;
    
        for (int n = 0; n < depth; n++)
        {
            series[k].mMean += series[k].mPoints[n];
        }
        
        series[k].mMean /= depth;
    }

    for (int k = 0; k < order; k++)
    {
        for (int n = k; n < order; n++)
        {
            covs[k].mPoints[n] = 0;

            for (int i = 0; i < depth; i++)
            {
                covs[k].mPoints[n] += (series[k].mPoints[i] - series[k].mMean) * (series[n].mPoints[i] - series[n].mMean);
            }

            covs[k].mPoints[n] = covs[k].mPoints[n] / (depth - n);
            covs[n].mPoints[k] = covs[k].mPoints[n];
        }
    }
}

void GetCorMatrix(SSeries& series[], SSeries& cors[], const int order)
{
    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            cors[k].mPoints[n] = series[k].mPoints[n] / MathSqrt(series[k].mPoints[k] * series[n].mPoints[n]);
        }
    }
}

double GetMaxInLuMatrix(SSeries& series[], int& X, int& Y, const int order)
{
    double max = 0;
    
    for (int k = 0; k < order; k++)
    {
        for (int n = k + 1; n < order; n++)
        {
            double candidate = MathAbs(series[k].mPoints[n]);
            
            if (candidate > max)
            {
                max = candidate;
                X = k;
                Y = n;
            }
        }
    }
    
    return max;
}

void MultiplyMatrix(SSeries& series[], SSeries& A[], SSeries& B[], const int order, const bool transposeFirst)
{
    ClearMatrix(iResults, order, order);

    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            iResults[k].mPoints[n] = 0;

            for (int i = 0; i < order; i++) 
            {
                iResults[k].mPoints[n] += transposeFirst ? A[i].mPoints[k] * B[i].mPoints[n] : A[k].mPoints[i] * B[i].mPoints[n];
            }
        }
    }
    
    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            series[k].mPoints[n] = iResults[k].mPoints[n];
        }
    }
}

void GetLuMatrix(SSeries& series[], SSeries& L[], SSeries& U[], const int order) 
{
    double summary = 0;

    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            summary = 0;
            
            for (int i = 0; i < k; i++)
            {
                summary += U[i].mPoints[n] * L[k].mPoints[i];
            }
            
            U[k].mPoints[n] = series[k].mPoints[n] - summary;
            L[k].mPoints[n] = U[n].mPoints[n] == 0 ? 1 : U[k].mPoints[n] / U[n].mPoints[n];
        }
    }
    
    for (int k = 0; k < order; k++)
    {
        for (int n = k + 1; n < order; n++)
        {
            U[n].mPoints[k] = 0;
            L[k].mPoints[n] = 0;
        }
    }
}

double GetDeterminant(SSeries& series[], const int order)
{
    if (order < 2)
    {
        return(series[0].mPoints[0]);
    }

    ClearMatrix(iUpperMatrix, order, order);
    ClearMatrix(iLowerMatrix, order, order);
    GetLuMatrix(series, iLowerMatrix, iUpperMatrix, order);

    double determinant = 1;

    for (int k = 0; k < order; k++)
    {
        determinant *= iUpperMatrix[k].mPoints[k];
    }
    
    return(determinant);
}

void GetMinor(SSeries& series[], SSeries& minor[], const int rowIndex, const int colIndex, const int order) 
{
    int row = 0;
    for (int k = 0; k < order; k++)
    {
        if (k != rowIndex)
        {
            int col = 0;
            for (int n = 0; n < order; n++)
            {
                if (n != colIndex)
                {
                    minor[row].mPoints[col++] = series[k].mPoints[n];
                }
            }
            row++;
        }
    }
}

double GetInverseMatrix(SSeries& series[], SSeries& inversion[], const int order)
{
    double determinant = 1 / GetDeterminant(series, order);

    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            ClearMatrix(iMinors, order, order);
            GetMinor(series, iMinors, k, n, order);
            inversion[k].mPoints[n] = ((k + n) % 2 == 1 ? -1 : 1) * determinant * GetDeterminant(iMinors, order - 1);
        }
    }
    
    return(1);
}

double RotateMatrix(SSeries& series[], SSeries& rotations[], int X, int Y, const int order)
{
    double range = series[X].mPoints[X] == series[Y].mPoints[Y] ? 1 : series[X].mPoints[X] - series[Y].mPoints[Y];

    ClearMatrix(iMultipliers, order, order);
    
    double angle = MathArctan(2 * series[X].mPoints[Y] / range) / 2;
    double angleSin = MathSin(angle);
    double angleCos = MathCos(angle);

    for (int k = 0; k < order; k++) 
    {
        ArrayFill(rotations[k].mPoints, 0, order, 0);
        rotations[k].mPoints[k] = iMultipliers[k].mPoints[k] = (k == X || k == Y) ? angleCos : 1;
    }
    
    rotations[Y].mPoints[X] = iMultipliers[Y].mPoints[X] =  angleSin;
    rotations[X].mPoints[Y] = iMultipliers[X].mPoints[Y] = -angleSin;
    
    MultiplyMatrix(series, iMultipliers, series, order, true);
    MultiplyMatrix(series, series, iMultipliers, order, false);
    
    return(1);
}

void GetEigenValues(SSeries& series[], SSeries& eigens[], SSeries& vectors[], const int order)
{
    double precision = 0.00001;
    int row = 0, col = 0, iterations = 1000;

    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            eigens[k].mPoints[n] = series[k].mPoints[n];
        }
    }

    ClearMatrix(vectors, order, order);
    ClearMatrix(iRotations, order, order);
    RotateMatrix(eigens, vectors, row, col, order);

    for (int k = 0; k < iterations; k++)
    {
        if (GetMaxInLuMatrix(eigens, row, col, order) < precision) 
        {
            break;
        }
        
        RotateMatrix(eigens, iRotations, row, col, order);
        MultiplyMatrix(vectors, vectors, iRotations, order, false);
    }
}

int GetOptimalVector(SSeries& eigens[], const int order, const bool minimize)
{
    int index = 0;
    double issue = minimize ? 1000000 : 0;

    for (int k = 0; k < order; k++) 
    {
        double eigen = MathAbs(eigens[k].mPoints[k]);
        bool condition = minimize ? eigen < issue : eigen > issue;
    
        if (condition)
        {
            issue = eigen;
            index = k;
        }
    }
    
    return(index);
}

double GetIndexValues(SSeries& series[], SSeries& vectors[], SSeries& results[], const int order, const int depth, const int index)
{
    double summary = 0;

    for (int k = 0; k < order; k++) 
    {
        summary += vectors[k].mPoints[index];
    }

    for (int n = 0; n < depth; n++) 
    {
        double product = 1;
    
        for (int k = 0; k < order; k++) 
        {
            product *= MathPow(series[k].mPoints[n], vectors[k].mPoints[index]);
        }

        results[0].mPoints[n] = MathLog(MathPow(product, 1.0 / summary));
    }

    return(index);
}

void GetEstimator(SSeries& series[], SSeries& covs[], SSeries& estimators[], const int order, const int depth)
{
    double R = 0, A = 0, B = 0, C = 0, summaryAkk = 0;
    
    // Shrink Target
    
    for (int k = 0; k < order - 1; k++) 
    {
        double summary = 0;
        
        for (int n = k + 1; n < order; n++) 
        {
            summary += covs[k].mPoints[n] / MathSqrt(covs[k].mPoints[k] * covs[n].mPoints[n]);
        }

        R += summary;
    }
    
    R = R * 2 / ((order - 1) * order);

    // Shrink Intensity

    for (int k = 0; k < order; k++) 
    {
        double summaryA = 0;
        
        for (int n = 0; n < order; n++) 
        {
            double Akn = 0;
            
            for (int i = 0; i < depth; i++)
            {
                double X = series[k].mPoints[i] - series[k].mMean;
                double Y = series[n].mPoints[i] - series[n].mMean;
                Akn += (X * Y - covs[k].mPoints[n]) * (X * Y - covs[k].mPoints[n]);
            }
            
            estimators[k].mPoints[n] = k == n ? covs[k].mPoints[n] : R * MathSqrt(covs[k].mPoints[k] * covs[n].mPoints[n]);
            summaryA += Akn / depth;
            summaryAkk += k == n ? summaryA : 0;
        }
        
        A += summaryA;
    }

    for (int k = 0; k < order; k++) 
    {
        double summaryB = 0, summaryC = 0;
        
        for (int n = 0; n < order; n++) 
        {
            double aCovKk = 0, aCovNn = 0;
            
            for (int i = 0; i < depth; i++)
            {
                double X = series[k].mPoints[i] - series[k].mMean;
                double Y = series[n].mPoints[i] - series[n].mMean;
                aCovKk += (X * X - covs[k].mPoints[k]) * (X * Y - covs[k].mPoints[n]);
                aCovNn += (Y * Y - covs[n].mPoints[n]) * (X * Y - covs[k].mPoints[n]);
            }
            
            summaryB += k == n ? 0 : (R / 2) * (MathSqrt(covs[n].mPoints[n] / covs[k].mPoints[k]) * aCovKk + MathSqrt(covs[k].mPoints[k] / covs[n].mPoints[n]) * aCovNn);
            summaryC += (estimators[k].mPoints[n] - series[k].mPoints[n]) * (estimators[k].mPoints[n] - series[k].mPoints[n]);
        }
        
        B += summaryB;
        C += summaryC;
    }
    
    B += summaryAkk;
    
    double intensity = MathMax(0, MathMin(1, ((A - B) / C) / depth));
    
    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            estimators[k].mPoints[n] = estimators[k].mPoints[n] * intensity + (1 - intensity) * covs[k].mPoints[n];
        }
    }
}

void ShowVector(SSeries& series[], SSeries& vectors[], const int order, const int index)
{
    double max = 0;    
    datetime times[];
    int X = 10, Y = 5, size = 8, precision = 5, offset = int(ChartGetDouble(0, CHART_SHIFT_SIZE) * ChartGetInteger(0, CHART_WIDTH_IN_PIXELS) / 100);

    for (int k = 0; k < order; k++) 
    {
        max = MathAbs(vectors[k].mPoints[index]) > max ? MathAbs(vectors[k].mPoints[index]) : max;
    }

    int maxPixel = offset - 155;

    for (int k = 0; k < order; k++) 
    {
        string alignment = (vectors[k].mPoints[index] > 0 ? " " : "") + DoubleToString(vectors[k].mPoints[index], precision);
        string label = StringFormat("%s  | %s", series[k].mName, alignment);
        int level = int(MathCeil(MathAbs(vectors[k].mPoints[index]) * maxPixel / max));

        ShowItem(OBJ_LABEL, series[k].mName + "A", label, X, Y, size, clrBlack);
        ShowItem(OBJ_RECTANGLE_LABEL, series[k].mName + "B", series[k].mName, X + level + 105, Y + 2, size, vectors[k].mPoints[index] > 0 ? clrDeepSkyBlue : clrTomato, level, 10);
        
        Y += size * 2;
    }
}

void ShowItem(const ENUM_OBJECT type, const string name, const string value, const int X, const int Y, const int size, const color clr, const int W = 0, const int H = 0, ENUM_ANCHOR_POINT point = ANCHOR_RIGHT_UPPER, ENUM_BASE_CORNER corner = CORNER_RIGHT_UPPER)
{
    long chart = ChartID();
    int window = ChartWindowFind();

    if (ObjectFind(chart, name) < 0)
    {
        ObjectCreate(chart, name, type, window, 0, 0);
    }

    ObjectSetInteger(chart, name, OBJPROP_ANCHOR, point);
    ObjectSetInteger(chart, name, OBJPROP_CORNER, corner);
    ObjectSetInteger(chart, name, OBJPROP_FONTSIZE, size);
    ObjectSetInteger(chart, name, OBJPROP_XDISTANCE, X);
    ObjectSetInteger(chart, name, OBJPROP_YDISTANCE, Y);
    ObjectSetInteger(chart, name, OBJPROP_COLOR, clr);
    ObjectSetString(chart, name, OBJPROP_TEXT, value);
    
    if (type == OBJ_RECTANGLE_LABEL) 
    {
        ObjectSetInteger(chart, name, OBJPROP_BGCOLOR, clr);
        ObjectSetInteger(chart, name, OBJPROP_BACK, false);
        ObjectSetInteger(chart, name, OBJPROP_FILL, true);
        ObjectSetInteger(chart, name, OBJPROP_WIDTH, 0);
        ObjectSetInteger(chart, name, OBJPROP_YSIZE, H);
        ObjectSetInteger(chart, name, OBJPROP_XSIZE, W);
    }
}

bool SetLine(const string name, const datetime time)
{
    long chart = ChartID();
    int window = ChartWindowFind();

    if (ObjectFind(chart, name) < 0)
    {
        ObjectCreate(chart, name, OBJ_VLINE, window, 0, 0);
    }
    
    ObjectSetInteger(chart, name, OBJPROP_RAY, true);
    ObjectSetInteger(chart, name, OBJPROP_COLOR, clrRed);
    ObjectSetInteger(chart, name, OBJPROP_STYLE, STYLE_DOT);
    
    ObjectMove(chart, name, 0, time, 0);
    
    return true;
}

void HidePanel()
{
    long chart = ChartID();
    int window = ChartWindowFind();
    ObjectsDeleteAll(chart, window, OBJ_VLINE);
    ObjectsDeleteAll(chart, window, OBJ_LABEL);
    ObjectsDeleteAll(chart, window, OBJ_RECTANGLE_LABEL);
}

void CustomLog(const string message)
{
    ShowItem(OBJ_LABEL, "Message", message, 10, 10, 10, clrRed, 0, 0, ANCHOR_LEFT_LOWER, CORNER_LEFT_LOWER);
}

void DebugMatrix(SSeries& series[])
{
    string result = "\n";
    int order = ArraySize(series);
    
    for (int i = 0; i < order; i++)
    {
       for (int j = 0; j < order; j++)
       {
           result += DoubleToString(series[i].mPoints[j], 5) + " ";
       }
       
       result += "\n";
    }
    
    Comment(result);
}